#
#  player_controller.rb
#  MazeWarper
#
#  Created by Mayur Pawashe on 8/30/08.
#  Copyright (c) 2008 zgcoder. All rights reserved.
#

require 'player'

class PlayerController
  include Math
  
  PLAYER_MOVE_SPEED = 0.01
  #the lower number, the faster mouse speed is
  MOUSE_SPEED = 7.0
  
  attr_reader :player
  
  def initialize
    @player = Player.new(2.0, 2.0, -1.0)
    @keys = []
    @mouse_inversion = -1.0
  end
  
  def key_down(key)
    @keys[key] = true
    
    if SDL::Key::T == key
      if @mouse_inversion < 0
        @mouse_inversion = 1.0
      elsif @mouse_inversion > 0
        @mouse_inversion = -1.0
      end
    end
  end
  
  def key_up(key)
    @keys[key] = false
  end
  
  def mouse_moved(x, y)
    #ignore first call of mouse_moved
    if @initiated
      if x > 0
        #turn right
        adjust_yaw(-1.0, x / MOUSE_SPEED)
      elsif x < 0
        #turn left
		adjust_yaw(-1.0, x / MOUSE_SPEED)
      end
    
      if y > 0
		    #look up
        new_pitch = @player.pitch + (-@mouse_inversion * (y / MOUSE_SPEED))
        
        @player.pitch = new_pitch if new_pitch > -90.0
      elsif y < 0
		    #look down
        new_pitch = @player.pitch + (@mouse_inversion * (-y / MOUSE_SPEED))
        
        @player.pitch = new_pitch if new_pitch < 90.0
      end
    else
      @initiated = true
    end
  end
  
  def collisions_are_okay?(new_position)
    current_cell_of_player = $maze_controller.maze.current_cell_of_player(@player)
    
    return true unless current_cell_of_player
    
    if current_cell_of_player == $maze_controller.maze.last_cell
      @player = Player.new(2.0, 2.0, -1.0)
      $maze_controller = MazeController.new()
      old_cell_of_player = nil
      return false
    end
    
    @old_cell_of_player.marked = true if @old_cell_of_player and @old_cell_of_player != current_cell_of_player and !@old_cell_of_player.marked
    
    @old_cell_of_player = current_cell_of_player
    
    unless current_cell_of_player.left_wall.passage
      #check left wall
      if ($maze_controller.maze.cube_radius * 2 * current_cell_of_player.x - new_position.x).abs <= 0.5
        return false
      end
    else
      if current_cell_of_player.x > 0
        left_cell = $maze_controller.maze.left_neighbor(current_cell_of_player)
        return true unless left_cell

        #check left cell's top wall
        if !left_cell.top_wall.passage and ($maze_controller.maze.cube_radius * 2 * current_cell_of_player.x - new_position.x).abs <= 0.5 and ($maze_controller.maze.cube_radius * 2 * (current_cell_of_player.y + 1) - new_position.z.abs).abs <= 0.5
          return false
        end

        #check left cell's bottom wall
        if !left_cell.bottom_wall.passage and ($maze_controller.maze.cube_radius * 2 * current_cell_of_player.x - new_position.x).abs <= 0.5 and ($maze_controller.maze.cube_radius * 2 * current_cell_of_player.y - new_position.z.abs).abs <= 0.5
          return false
        end
      end
    end
    
    unless current_cell_of_player.right_wall.passage
      #check right wall
      if ($maze_controller.maze.cube_radius * 2 * (current_cell_of_player.x + 1) - new_position.x).abs <= 0.5
        return false
      end
    else
      if current_cell_of_player.x < $maze_controller.maze.size - 1
        right_cell = $maze_controller.maze.right_neighbor(current_cell_of_player)
        return true unless right_cell
    
        #check right cell's top wall
        if !right_cell.top_wall.passage and ($maze_controller.maze.cube_radius * 2 * (current_cell_of_player.x + 1) - new_position.x).abs <= 0.5 and ($maze_controller.maze.cube_radius * 2 * (current_cell_of_player.y + 1) - new_position.z.abs).abs <= 0.5
          return false
        end
    
        #check right cell's bottom wall
        if !right_cell.bottom_wall.passage and ($maze_controller.maze.cube_radius * 2 * (current_cell_of_player.x + 1) - new_position.x).abs <= 0.5 and ($maze_controller.maze.cube_radius * 2 * current_cell_of_player.y - new_position.z.abs).abs <= 0.5
          return false
        end
      end
    end
    
    unless current_cell_of_player.bottom_wall.passage
      #check bottom wall
      if ($maze_controller.maze.cube_radius * 2 * current_cell_of_player.y - new_position.z.abs).abs <= 0.5
        return false
      end
    else
      if current_cell_of_player.y > 0
        bottom_cell = $maze_controller.maze.bottom_neighbor(current_cell_of_player)
        return true unless bottom_cell
        
        #check bottom cell's right wall
        if !bottom_cell.right_wall.passage and ($maze_controller.maze.cube_radius * 2 * (current_cell_of_player.x + 1) - new_position.x).abs <= 0.5 and ($maze_controller.maze.cube_radius * 2 * current_cell_of_player.y - new_position.z.abs).abs <= 0.5
          return false
        end
        
        #check bottom cell's left wall
        if !bottom_cell.left_wall.passage and ($maze_controller.maze.cube_radius * 2 * current_cell_of_player.x - new_position.x).abs <= 0.5 and ($maze_controller.maze.cube_radius * 2 * current_cell_of_player.y - new_position.z.abs).abs <= 0.5
          return false
        end
      end
    end
    
    unless current_cell_of_player.top_wall.passage
      #check top wall
      if ($maze_controller.maze.cube_radius * 2 * (current_cell_of_player.y + 1) - new_position.z.abs).abs <= 0.5
        return false
      end
    else
      top_cell = $maze_controller.maze.top_neighbor(current_cell_of_player)
      return true unless top_cell
      
      #check top cell's right wall
      if !top_cell.right_wall.passage and ($maze_controller.maze.cube_radius * 2 * (current_cell_of_player.x + 1) - new_position.x).abs <= 0.5 and ($maze_controller.maze.cube_radius * 2 * (current_cell_of_player.y + 1) - new_position.z.abs).abs <= 0.5
        return false
      end
      
      #check top cell's left wall
      if !top_cell.left_wall.passage and ($maze_controller.maze.cube_radius * 2 * current_cell_of_player.x - new_position.x).abs <= 0.5 and ($maze_controller.maze.cube_radius * 2 * (current_cell_of_player.y + 1) - new_position.z.abs).abs <= 0.5
        return false
      end
    end
    
    true
  end
  
  #only moves forward or backward
  def move_at_direction(direction)
    new_position = Vector.new
		
		new_position.x = @player.position.x + (@player.direction.x * PLAYER_MOVE_SPEED) * direction
		new_position.y = @player.position.y + (@player.direction.y * PLAYER_MOVE_SPEED) * direction
		new_position.z = @player.position.z + (@player.direction.z * PLAYER_MOVE_SPEED) * direction
		
		@player.position = new_position if collisions_are_okay?(new_position)
  end
  
  #angle is in degrees
  def move_at_angle(y_direction, vector_direction, angle)
    rotational_vector = Vector.new(0.0, y_direction, 0.0)
    
    quaternion = Quaternion.quaternion_from_axis_angle(rotational_vector, angle * (PI / 180.0))
    vector = Quaternion.multiply_by_vector(quaternion, @player.direction)
    
    new_position = Vector.new
    
    new_position.x = @player.position.x + (vector.x * PLAYER_MOVE_SPEED) * vector_direction
		new_position.y = @player.position.y + (vector.y * PLAYER_MOVE_SPEED) * vector_direction
		new_position.z = @player.position.z + (vector.z * PLAYER_MOVE_SPEED) * vector_direction
		
		@player.position = new_position if collisions_are_okay?(new_position)
  end
  
  def adjust_yaw(y_direction, angle)
    rotational_vector = Vector.new(0.0, y_direction, 0.0)
		
		quaternion = Quaternion.quaternion_from_axis_angle(rotational_vector, angle * (PI / 180.0))
		@player.direction = Quaternion.multiply_by_vector(quaternion, @player.direction)
		
		@player.yaw += angle
		
	  if @player.yaw >= 360.0
	    @player.yaw -= 360.0
	  elsif @player.yaw <= -360.0
	    @player.yaw += 360
	  end
  end
  
  def update
    if @keys[SDL::Key::W] and @keys[SDL::Key::A]
      #forward left
			move_at_angle(1.0, -1.0, 45.0)
		elsif @keys[SDL::Key::W] and @keys[SDL::Key::D]
		  #forward right
			move_at_angle(-1.0, -1.0, 45.0)
		elsif @keys[SDL::Key::S] and @keys[SDL::Key::A]
			#backward left
			move_at_angle(-1.0, 1.0, 45.0)
		elsif @keys[SDL::Key::S] and @keys[SDL::Key::D]
			#backward right
			move_at_angle(1.0, 1.0, 45.0)
		else
		  if @keys[SDL::Key::W]
				#forward
				move_at_direction(-1.0)
			end
			
			if @keys[SDL::Key::S]
				#backwards
				move_at_direction(1.0)
			end
			
			if @keys[SDL::Key::A]
				#sidestep left
  		  move_at_angle(-1.0, 1.0, 90.0)
			end
			
			if @keys[SDL::Key::D]
				#sidestep right
				move_at_angle(1.0, 1.0, 90.0)
			end
    end
    
    if @keys[SDL::Key::I]
			#look up
			if @player.pitch > -90.0
				@player.pitch -= 0.1
			end
		end
		
		if @keys[SDL::Key::K]
			#look down
			if @player.pitch < 90.0
				@player.pitch += 0.1
			end
		end
    
    if @keys[SDL::Key::J]
			#turn left
			adjust_yaw(-1.0, -0.1)
		end
		
		if @keys[SDL::Key::L]
			#turn right
			adjust_yaw(-1.0, 0.1)
		end
  end
end
