#
#  quaternion.rb
#  Boom
#
#  Created by Mayur Pawashe on 5/17/08.
#  Copyright (c) 2008 zgcoder. All rights reserved.
#

require 'vector'

class Quaternion
	attr_accessor :x, :y, :z, :w
	
	def initialize
		@x = 0.0
		@y = 0.0
		@z = 0.0
		@w = 1.0
	end
	
	def Quaternion.invert(quaternion)
		length = 1.0 / ((quaternion.x * quaternion.x) + (quaternion.y * quaternion.y) + (quaternion.z * quaternion.z) + (quaternion.w * quaternion.w))
		
		inverted_quaternion = Quaternion.new
		inverted_quaternion.x = quaternion.x * -length
		inverted_quaternion.y = quaternion.y * -length
		inverted_quaternion.z = quaternion.z * -length
		inverted_quaternion.w = quaternion.w * length
		
		return inverted_quaternion
	end
	
	def Quaternion.multiply(quaternion1, quaternion2)
		vector1 = Vector.new
		vector1.x = quaternion1.x
		vector1.y = quaternion1.y
		vector1.z = quaternion1.z
		
		vector2 = Vector.new
		vector2.x = quaternion2.x
		vector2.y = quaternion2.y
		vector2.z = quaternion2.z
		
		angle = quaternion1.w * quaternion2.w - Vector.dot_product(vector1, vector2)
		
		cross = Vector.cross_product(vector1, vector2)
		
		vector1.x *= quaternion2.w;
		vector1.y *= quaternion2.w;
		vector1.z *= quaternion2.w;
		
		vector2.x *= quaternion1.w;
		vector2.y *= quaternion1.w;
		vector2.z *= quaternion1.w;
		
		multiplied_quaternion = Quaternion.new
		
		multiplied_quaternion.x = vector1.x + vector2.x + cross.x;
		multiplied_quaternion.y = vector1.y + vector2.y + cross.y;
		multiplied_quaternion.z = vector1.z + vector2.z + cross.z;
		multiplied_quaternion.w = angle;
		
		return multiplied_quaternion
	end
	
	def Quaternion.multiply_by_vector(quaternion, vector)
		vector_quaternion = Quaternion.new
		vector_quaternion.x = vector.x
		vector_quaternion.y = vector.y
		vector_quaternion.z = vector.z
		vector_quaternion.w = 0.0
		
		inverse_quaternion = Quaternion.invert(quaternion)
		result_quaternion = Quaternion.multiply(vector_quaternion, inverse_quaternion)
		result_quaternion = Quaternion.multiply(quaternion, result_quaternion)
		
		result_vector = Quaternion.new
		
		result_vector.x = result_quaternion.x
		result_vector.y = result_quaternion.y
		result_vector.z = result_quaternion.z
		
		return result_vector
	end
	
	def Quaternion.quaternion_from_axis_angle(axis_vector, anAngle)
		angle = anAngle * 0.5
		axis_vector = Vector.normalize(axis_vector)
		sinAngle = Math.sin(angle)
		
		quaternion = Quaternion.new
		quaternion.x = axis_vector.x * sinAngle
		quaternion.y = axis_vector.y * sinAngle
		quaternion.z = axis_vector.z * sinAngle
		quaternion.w = Math.cos(angle)
		
		return quaternion
	end
end
